/*
 * Automatically generated C config: don't edit
 */
#define AUTOCONF_INCLUDED

/*
 * Code maturity level options
 */
#define CONFIG_EXPERIMENTAL 1

/*
 * Loadable module support
 */
#define CONFIG_MODULES 1
#define CONFIG_MODVERSIONS 1
#define CONFIG_KMOD 1

/*
 * General setup
 */
#define CONFIG_VT 1
#define CONFIG_VT_CONSOLE 1
#define CONFIG_SPARC32 1
#undef  CONFIG_AP1000
#undef  CONFIG_SMP
#undef  CONFIG_SUN4
#undef  CONFIG_PCI

/*
 * Console drivers
 */
#define CONFIG_PROM_CONSOLE 1
#define CONFIG_FB 1
#define CONFIG_DUMMY_CONSOLE 1
#define CONFIG_FB_SBUS 1
#define CONFIG_FB_CGSIX 1
#define CONFIG_FB_BWTWO 1
#define CONFIG_FB_CGTHREE 1
#define CONFIG_FB_TCX 1
#define CONFIG_FB_CGFOURTEEN 1
#define CONFIG_FB_LEO 1
#undef  CONFIG_FB_VIRTUAL
#undef  CONFIG_FBCON_ADVANCED
#define CONFIG_FBCON_MFB 1
#define CONFIG_FBCON_CFB8 1
#define CONFIG_FBCON_FONTWIDTH8_ONLY 1
#define CONFIG_FONT_SUN8x16 1
#undef  CONFIG_FBCON_FONTS
#define CONFIG_SBUS 1
#define CONFIG_SBUSCHAR 1
#define CONFIG_SUN_MOUSE 1
#define CONFIG_SERIAL 1
#define CONFIG_SUN_SERIAL 1
#define CONFIG_SERIAL_CONSOLE 1
#define CONFIG_SUN_KEYBOARD 1
#define CONFIG_SUN_CONSOLE 1
#define CONFIG_SUN_AUXIO 1
#define CONFIG_SUN_IO 1

/*
 * Misc Linux/SPARC drivers
 */
#undef  CONFIG_SUN_OPENPROMIO
#define CONFIG_SUN_OPENPROMIO_MODULE 1
#define CONFIG_SUN_MOSTEK_RTC 1
#undef  CONFIG_SUN_BPP
#undef  CONFIG_SUN_VIDEOPIX
#undef  CONFIG_SUN_AURORA
#define CONFIG_SUN_AURORA_MODULE 1

/*
 * Linux/SPARC audio subsystem (EXPERIMENTAL)
 */
#undef  CONFIG_SPARCAUDIO
#undef  CONFIG_SPARCAUDIO_AMD7930
#undef  CONFIG_SPARCAUDIO_DBRI
#undef  CONFIG_SPARCAUDIO_CS4231
#undef  CONFIG_SPARCAUDIO_DUMMY
#undef  CONFIG_SUN_OPENPROMFS
#define CONFIG_SUN_OPENPROMFS_MODULE 1
#define CONFIG_NET 1
#define CONFIG_SYSVIPC 1
#undef  CONFIG_BSD_PROCESS_ACCT
#define CONFIG_SYSCTL 1
#define CONFIG_BINFMT_AOUT 1
#define CONFIG_BINFMT_ELF 1
#undef  CONFIG_BINFMT_MISC
#define CONFIG_BINFMT_MISC_MODULE 1
#undef  CONFIG_BINFMT_JAVA
#define CONFIG_BINFMT_JAVA_MODULE 1
#define CONFIG_SUNOS_EMUL 1

/*
 * Floppy, IDE, and other block devices
 */
#define CONFIG_BLK_DEV_FD 1
#define CONFIG_BLK_DEV_MD 1
#undef  CONFIG_MD_LINEAR
#define CONFIG_MD_LINEAR_MODULE 1
#undef  CONFIG_MD_STRIPED
#define CONFIG_MD_STRIPED_MODULE 1
#undef  CONFIG_MD_MIRRORING
#define CONFIG_MD_MIRRORING_MODULE 1
#undef  CONFIG_MD_RAID5
#define CONFIG_MD_RAID5_MODULE 1
#define CONFIG_BLK_DEV_RAM 1
#define CONFIG_BLK_DEV_RAM_SIZE (4096)
#define CONFIG_BLK_DEV_INITRD 1
#undef  CONFIG_BLK_DEV_LOOP
#define CONFIG_BLK_DEV_LOOP_MODULE 1
#undef  CONFIG_BLK_DEV_NBD
#define CONFIG_BLK_DEV_NBD_MODULE 1

/*
 * Networking options
 */
#define CONFIG_PACKET 1
#undef  CONFIG_NETLINK
#undef  CONFIG_FIREWALL
#undef  CONFIG_FILTER
#define CONFIG_UNIX 1
#define CONFIG_INET 1
#undef  CONFIG_IP_MULTICAST
#undef  CONFIG_IP_ADVANCED_ROUTER
#undef  CONFIG_IP_PNP
#undef  CONFIG_IP_ROUTER
#undef  CONFIG_NET_IPIP
#undef  CONFIG_NET_IPGRE
#undef  CONFIG_IP_ALIAS
#undef  CONFIG_SYN_COOKIES

/*
 * (it is safe to leave these untouched)
 */
#undef  CONFIG_INET_RARP
#define CONFIG_INET_RARP_MODULE 1
#define CONFIG_SKB_LARGE 1
#undef  CONFIG_IPV6
#define CONFIG_IPV6_MODULE 1
#undef  CONFIG_IPV6_EUI64

/*
 *  
 */
#undef  CONFIG_IPX
#define CONFIG_IPX_MODULE 1
#undef  CONFIG_IPX_INTERN
#undef  CONFIG_SPX
#undef  CONFIG_ATALK
#define CONFIG_ATALK_MODULE 1
#undef  CONFIG_X25
#undef  CONFIG_LAPB
#undef  CONFIG_BRIDGE
#undef  CONFIG_NET_DIVERT
#undef  CONFIG_LLC
#undef  CONFIG_ECONET
#undef  CONFIG_WAN_ROUTER
#undef  CONFIG_NET_FASTROUTE
#undef  CONFIG_NET_HW_FLOWCONTROL
#undef  CONFIG_CPU_IS_SLOW

/*
 * QoS and/or fair queueing
 */
#undef  CONFIG_NET_SCHED

/*
 * Amateur Radio support
 */
#undef  CONFIG_HAMRADIO

/*
 * ISDN subsystem
 */
#undef  CONFIG_ISDN

/*
 * SCSI support
 */
#define CONFIG_SCSI 1

/*
 * SCSI support type (disk, tape, CDrom)
 */
#define CONFIG_BLK_DEV_SD 1
#define CONFIG_CHR_DEV_ST 1
#define CONFIG_BLK_DEV_SR 1
#define CONFIG_BLK_DEV_SR_VENDOR 1
#undef  CONFIG_CHR_DEV_SG
#define CONFIG_CHR_DEV_SG_MODULE 1

/*
 * Some SCSI devices (e.g. CD jukebox) support multiple LUNs
 */
#define CONFIG_SCSI_MULTI_LUN 1
#define CONFIG_SCSI_CONSTANTS 1

/*
 * SCSI low-level drivers
 */
#define CONFIG_SCSI_SUNESP 1
#undef  CONFIG_SCSI_QLOGICPTI
#define CONFIG_SCSI_QLOGICPTI_MODULE 1

/*
 * Fibre Channel support
 */
#undef  CONFIG_FC4
#define CONFIG_FC4_MODULE 1

/*
 * FC4 drivers
 */
#undef  CONFIG_FC4_SOC
#define CONFIG_FC4_SOC_MODULE 1
#undef  CONFIG_FC4_SOCAL
#define CONFIG_FC4_SOCAL_MODULE 1

/*
 * FC4 targets
 */
#undef  CONFIG_SCSI_PLUTO
#define CONFIG_SCSI_PLUTO_MODULE 1
#undef  CONFIG_SCSI_FCAL
#define CONFIG_SCSI_FCAL_MODULE 1

/*
 * Network device support
 */
#define CONFIG_NETDEVICES 1
#undef  CONFIG_DUMMY
#define CONFIG_DUMMY_MODULE 1
#undef  CONFIG_BONDING
#define CONFIG_BONDING_MODULE 1
#undef  CONFIG_PPP
#define CONFIG_PPP_MODULE 1

/*
 * CCP compressors for PPP are only built as modules.
 */
#undef  CONFIG_SLIP
#define CONFIG_SLIP_MODULE 1
#define CONFIG_SLIP_COMPRESSED 1
#define CONFIG_SLIP_SMART 1
#undef  CONFIG_SLIP_MODE_SLIP6
#define CONFIG_SUNLANCE 1
#undef  CONFIG_HAPPYMEAL
#define CONFIG_HAPPYMEAL_MODULE 1
#undef  CONFIG_SUNBMAC
#define CONFIG_SUNBMAC_MODULE 1
#undef  CONFIG_SUNQE
#define CONFIG_SUNQE_MODULE 1
#undef  CONFIG_MYRI_SBUS
#define CONFIG_MYRI_SBUS_MODULE 1

/*
 * Unix98 PTY support
 */
#define CONFIG_UNIX98_PTYS 1
#define CONFIG_UNIX98_PTY_COUNT (256)

/*
 * Filesystems
 */
#undef  CONFIG_QUOTA
#undef  CONFIG_AUTOFS_FS
#define CONFIG_AUTOFS_FS_MODULE 1
#undef  CONFIG_ADFS_FS
#undef  CONFIG_AFFS_FS
#define CONFIG_AFFS_FS_MODULE 1
#undef  CONFIG_HFS_FS
#undef  CONFIG_FAT_FS
#define CONFIG_FAT_FS_MODULE 1
#undef  CONFIG_MSDOS_FS
#define CONFIG_MSDOS_FS_MODULE 1
#undef  CONFIG_UMSDOS_FS
#undef  CONFIG_VFAT_FS
#define CONFIG_VFAT_FS_MODULE 1
#undef  CONFIG_ISO9660_FS
#define CONFIG_ISO9660_FS_MODULE 1
#undef  CONFIG_JOLIET
#undef  CONFIG_MINIX_FS
#define CONFIG_MINIX_FS_MODULE 1
#undef  CONFIG_NTFS_FS
#undef  CONFIG_HPFS_FS
#define CONFIG_HPFS_FS_MODULE 1
#define CONFIG_PROC_FS 1
#define CONFIG_DEVPTS_FS 1
#undef  CONFIG_QNX4FS_FS
#undef  CONFIG_ROMFS_FS
#define CONFIG_ROMFS_FS_MODULE 1
#define CONFIG_EXT2_FS 1
#undef  CONFIG_SYSV_FS
#define CONFIG_SYSV_FS_MODULE 1
#undef  CONFIG_UFS_FS
#define CONFIG_UFS_FS_MODULE 1
#define CONFIG_UFS_FS_WRITE 1
#undef  CONFIG_EFS_FS
#define CONFIG_EFS_FS_MODULE 1
#define CONFIG_SGI_PARTITION 1

/*
 * Network File Systems
 */
#undef  CONFIG_CODA_FS
#define CONFIG_CODA_FS_MODULE 1
#define CONFIG_NFS_FS 1
#define CONFIG_NFS_V3 1
#undef  CONFIG_NFSD
#define CONFIG_NFSD_MODULE 1
#define CONFIG_NFSD_V3 1
#undef  CONFIG_NFSD_TCP
#define CONFIG_SUNRPC 1
#define CONFIG_LOCKD 1
#undef  CONFIG_SMB_FS
#define CONFIG_SMB_FS_MODULE 1
#undef  CONFIG_SMB_NLS_DEFAULT
#undef  CONFIG_NCP_FS
#define CONFIG_NCP_FS_MODULE 1
#undef  CONFIG_NCPFS_PACKET_SIGNING
#undef  CONFIG_NCPFS_IOCTL_LOCKING
#undef  CONFIG_NCPFS_STRONG
#undef  CONFIG_NCPFS_NFS_NS
#undef  CONFIG_NCPFS_OS2_NS
#undef  CONFIG_NCPFS_MOUNT_SUBDIR
#undef  CONFIG_NCPFS_NLS
#undef  CONFIG_NCPFS_EXTRAS

/*
 * Partition Types
 */
#define CONFIG_BSD_DISKLABEL 1
#undef  CONFIG_MAC_PARTITION
#undef  CONFIG_MINIX_SUBPARTITION
#define CONFIG_SMD_DISKLABEL 1
#define CONFIG_SOLARIS_X86_PARTITION 1
#undef  CONFIG_UNIXWARE_DISKLABEL
#define CONFIG_AMIGA_PARTITION 1
#define CONFIG_NLS 1

/*
 * Native Language Support
 */
#define CONFIG_NLS_DEFAULT "cp437"
#undef  CONFIG_NLS_CODEPAGE_437
#undef  CONFIG_NLS_CODEPAGE_737
#undef  CONFIG_NLS_CODEPAGE_775
#undef  CONFIG_NLS_CODEPAGE_850
#undef  CONFIG_NLS_CODEPAGE_852
#undef  CONFIG_NLS_CODEPAGE_855
#undef  CONFIG_NLS_CODEPAGE_857
#undef  CONFIG_NLS_CODEPAGE_860
#undef  CONFIG_NLS_CODEPAGE_861
#undef  CONFIG_NLS_CODEPAGE_862
#undef  CONFIG_NLS_CODEPAGE_863
#undef  CONFIG_NLS_CODEPAGE_864
#undef  CONFIG_NLS_CODEPAGE_865
#undef  CONFIG_NLS_CODEPAGE_866
#undef  CONFIG_NLS_CODEPAGE_869
#undef  CONFIG_NLS_CODEPAGE_874
#undef  CONFIG_NLS_CODEPAGE_932
#undef  CONFIG_NLS_CODEPAGE_936
#undef  CONFIG_NLS_CODEPAGE_949
#undef  CONFIG_NLS_CODEPAGE_950
#undef  CONFIG_NLS_ISO8859_1
#undef  CONFIG_NLS_ISO8859_2
#undef  CONFIG_NLS_ISO8859_3
#undef  CONFIG_NLS_ISO8859_4
#undef  CONFIG_NLS_ISO8859_5
#undef  CONFIG_NLS_ISO8859_6
#undef  CONFIG_NLS_ISO8859_7
#undef  CONFIG_NLS_ISO8859_8
#undef  CONFIG_NLS_ISO8859_9
#undef  CONFIG_NLS_ISO8859_14
#undef  CONFIG_NLS_ISO8859_15
#undef  CONFIG_NLS_KOI8_R
#undef  CONFIG_NLS_KOI8_RU

/*
 * Watchdog
 */
#undef  CONFIG_SOFT_WATCHDOG

/*
 * Kernel hacking
 */
#undef  CONFIG_MAGIC_SYSRQ
