/*
 * include/linux/m68kserial.h
 *
 * Copyright (C) 1992 by Theodore Ts'o.
 * 
 * Redistribution of this file is permitted under the terms of the GNU 
 * Public License (GPL)
 */

#ifndef _LINUX_M68KSERIAL_H
#define _LINUX_M68KSERIAL_H


#include <asm/m68kserial.h>

#endif /* _LINUX_M68KSERIAL_H*/
