/* $Id: isdn_ttyfax.h,v 1.2 2000/05/11 22:29:21 kai Exp $

 * header for Linux ISDN subsystem, tty_fax related functions (linklevel).
 *
 * Copyright 1999   by Armin Schindler (mac@melware.de)
 * Copyright 1999   by Ralf Spachmann (mel@melware.de)
 * Copyright 1999   by Cytronics & Melware
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#define XON	0x11
#define XOFF	0x13
#define DC2	0x12

