  /*
   * Copyright 1996 The Australian National University.
   * Copyright 1996 Fujitsu Laboratories Limited
   * 
   * This software may be distributed under the terms of the Gnu
   * Public License version 2 or later
  */
/* FDDI register pointers */
extern volatile struct formac *mac;
extern volatile struct plc *plc;
extern volatile int *csr0;
extern volatile int *csr1;
extern volatile int *buffer_mem;
extern volatile int *fifo;
