/* 
 * File...........: linux/drivers/s390/block/dasd_eckd_erp.h
 * Author(s)......: Holger Smolinski <Holger.Smolinski@de.ibm.com>
 * Bugreports.to..: <Linux390@de.ibm.com>
 * (C) IBM Corporation, IBM Deutschland Entwicklung GmbH, 2000
 */

#include <asm/dasd.h>
#include "dasd_erp.h"

#define PRINTK_HEADER "dasd_erp(eckd)"

dasd_era_t
dasd_eckd_erp_examine (cqr_t * cqr, devstat_t * stat)
{

}
