/******************************************************************************
 *
 *	(C)Copyright 1998,1999 SysKonnect,
 *	a business unit of Schneider & Koch & Co. Datensysteme GmbH.
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	The information in this file is provided "AS IS" without warranty.
 *
 ******************************************************************************/

/*
 * defines for all SMT attributes
 */

/*
 * this boring file was produced by perl
 * thanks Larry !
 */
#define	SMT_P0012	0x0012

#define	SMT_P0015	0x0015
#define	SMT_P0016	0x0016
#define	SMT_P0017	0x0017
#define	SMT_P0018	0x0018
#define	SMT_P0019	0x0019

#define	SMT_P001A	0x001a
#define	SMT_P001B	0x001b
#define	SMT_P001C	0x001c
#define	SMT_P001D	0x001d

#define	SMT_P100A	0x100a
#define	SMT_P100B	0x100b
#define	SMT_P100C	0x100c
#define	SMT_P100D	0x100d
#define	SMT_P100E	0x100e
#define	SMT_P100F	0x100f
#define	SMT_P1010	0x1010
#define	SMT_P1011	0x1011
#define	SMT_P1012	0x1012
#define	SMT_P1013	0x1013
#define	SMT_P1014	0x1014
#define	SMT_P1015	0x1015
#define	SMT_P1016	0x1016
#define	SMT_P1017	0x1017
#define	SMT_P1018	0x1018
#define	SMT_P1019	0x1019
#define	SMT_P101A	0x101a
#define	SMT_P101B	0x101b
#define	SMT_P101C	0x101c
#define	SMT_P101D	0x101d
#define	SMT_P101E	0x101e
#define	SMT_P101F	0x101f
#define	SMT_P1020	0x1020
#define	SMT_P1021	0x1021
#define	SMT_P1022	0x1022
#define	SMT_P1023	0x1023
#define	SMT_P1024	0x1024
#define	SMT_P1025	0x1025
#define	SMT_P1026	0x1026
#define	SMT_P1027	0x1027
#define	SMT_P1028	0x1028
#define	SMT_P1029	0x1029
#define	SMT_P102A	0x102a
#define	SMT_P102B	0x102b
#define	SMT_P102C	0x102c
#define	SMT_P102D	0x102d
#define	SMT_P102E	0x102e
#define	SMT_P102F	0x102f
#define	SMT_P1030	0x1030
#define	SMT_P1031	0x1031
#define	SMT_P1032	0x1032
#define	SMT_P1033	0x1033
#define	SMT_P1034	0x1034
#define	SMT_P1035	0x1035
#define	SMT_P1036	0x1036
#define	SMT_P1037	0x1037
#define	SMT_P1038	0x1038
#define	SMT_P1039	0x1039
#define	SMT_P103A	0x103a
#define	SMT_P103B	0x103b
#define	SMT_P103C	0x103c
#define	SMT_P103D	0x103d
#define	SMT_P103E	0x103e
#define	SMT_P103F	0x103f
#define	SMT_P1040	0x1040
#define	SMT_P1041	0x1041
#define	SMT_P1042	0x1042
#define	SMT_P1043	0x1043
#define	SMT_P1044	0x1044
#define	SMT_P1045	0x1045
#define	SMT_P1046	0x1046
#define	SMT_P1047	0x1047
#define	SMT_P1048	0x1048
#define	SMT_P1049	0x1049
#define	SMT_P104A	0x104a
#define	SMT_P104B	0x104b
#define	SMT_P104C	0x104c
#define	SMT_P104D	0x104d
#define	SMT_P104E	0x104e
#define	SMT_P104F	0x104f
#define	SMT_P1050	0x1050
#define	SMT_P1051	0x1051
#define	SMT_P1052	0x1052
#define	SMT_P1053	0x1053
#define	SMT_P1054	0x1054

#define	SMT_P10F0	0x10f0
#define	SMT_P10F1	0x10f1
#ifdef	ESS
#define	SMT_P10F2	0x10f2
#define	SMT_P10F3	0x10f3
#define	SMT_P10F4	0x10f4
#define	SMT_P10F5	0x10f5
#define	SMT_P10F6	0x10f6
#define	SMT_P10F7	0x10f7
#endif
#ifdef	SBA
#define	SMT_P10F8	0x10f8
#define	SMT_P10F9	0x10f9
#endif

#define	SMT_P200A	0x200a
#define	SMT_P200B	0x200b
#define	SMT_P200C	0x200c
#define	SMT_P200D	0x200d
#define	SMT_P200E	0x200e
#define	SMT_P200F	0x200f
#define	SMT_P2010	0x2010
#define	SMT_P2011	0x2011
#define	SMT_P2012	0x2012
#define	SMT_P2013	0x2013
#define	SMT_P2014	0x2014
#define	SMT_P2015	0x2015
#define	SMT_P2016	0x2016
#define	SMT_P2017	0x2017
#define	SMT_P2018	0x2018
#define	SMT_P2019	0x2019
#define	SMT_P201A	0x201a
#define	SMT_P201B	0x201b
#define	SMT_P201C	0x201c
#define	SMT_P201D	0x201d
#define	SMT_P201E	0x201e
#define	SMT_P201F	0x201f
#define	SMT_P2020	0x2020
#define	SMT_P2021	0x2021
#define	SMT_P2022	0x2022
#define	SMT_P2023	0x2023
#define	SMT_P2024	0x2024
#define	SMT_P2025	0x2025
#define	SMT_P2026	0x2026
#define	SMT_P2027	0x2027
#define	SMT_P2028	0x2028
#define	SMT_P2029	0x2029
#define	SMT_P202A	0x202a
#define	SMT_P202B	0x202b
#define	SMT_P202C	0x202c
#define	SMT_P202D	0x202d
#define	SMT_P202E	0x202e
#define	SMT_P202F	0x202f
#define	SMT_P2030	0x2030
#define	SMT_P2031	0x2031
#define	SMT_P2032	0x2032
#define	SMT_P2033	0x2033
#define	SMT_P2034	0x2034
#define	SMT_P2035	0x2035
#define	SMT_P2036	0x2036
#define	SMT_P2037	0x2037
#define	SMT_P2038	0x2038
#define	SMT_P2039	0x2039
#define	SMT_P203A	0x203a
#define	SMT_P203B	0x203b
#define	SMT_P203C	0x203c
#define	SMT_P203D	0x203d
#define	SMT_P203E	0x203e
#define	SMT_P203F	0x203f
#define	SMT_P2040	0x2040
#define	SMT_P2041	0x2041
#define	SMT_P2042	0x2042
#define	SMT_P2043	0x2043
#define	SMT_P2044	0x2044
#define	SMT_P2045	0x2045
#define	SMT_P2046	0x2046
#define	SMT_P2047	0x2047
#define	SMT_P2048	0x2048
#define	SMT_P2049	0x2049
#define	SMT_P204A	0x204a
#define	SMT_P204B	0x204b
#define	SMT_P204C	0x204c
#define	SMT_P204D	0x204d
#define	SMT_P204E	0x204e
#define	SMT_P204F	0x204f
#define	SMT_P2050	0x2050
#define	SMT_P2051	0x2051
#define	SMT_P2052	0x2052
#define	SMT_P2053	0x2053
#define	SMT_P2054	0x2054
#define	SMT_P2055	0x2055
#define	SMT_P2056	0x2056
#define	SMT_P2057	0x2057
#define	SMT_P2058	0x2058
#define	SMT_P2059	0x2059
#define	SMT_P205A	0x205a
#define	SMT_P205B	0x205b
#define	SMT_P205C	0x205c
#define	SMT_P205D	0x205d
#define	SMT_P205E	0x205e
#define	SMT_P205F	0x205f
#define	SMT_P2060	0x2060
#define	SMT_P2061	0x2061
#define	SMT_P2062	0x2062
#define	SMT_P2063	0x2063
#define	SMT_P2064	0x2064
#define	SMT_P2065	0x2065
#define	SMT_P2066	0x2066
#define	SMT_P2067	0x2067
#define	SMT_P2068	0x2068
#define	SMT_P2069	0x2069
#define	SMT_P206A	0x206a
#define	SMT_P206B	0x206b
#define	SMT_P206C	0x206c
#define	SMT_P206D	0x206d
#define	SMT_P206E	0x206e
#define	SMT_P206F	0x206f
#define	SMT_P2070	0x2070
#define	SMT_P2071	0x2071
#define	SMT_P2072	0x2072
#define	SMT_P2073	0x2073
#define	SMT_P2074	0x2074
#define	SMT_P2075	0x2075
#define	SMT_P2076	0x2076

#define	SMT_P208C	0x208c
#define	SMT_P208D	0x208d
#define	SMT_P208E	0x208e
#define	SMT_P208F	0x208f
#define	SMT_P2090	0x2090

#define	SMT_P20F0	0x20F0
#define	SMT_P20F1	0x20F1

#define	SMT_P320A	0x320a
#define	SMT_P320B	0x320b
#define	SMT_P320C	0x320c
#define	SMT_P320D	0x320d
#define	SMT_P320E	0x320e
#define	SMT_P320F	0x320f
#define	SMT_P3210	0x3210
#define	SMT_P3211	0x3211
#define	SMT_P3212	0x3212
#define	SMT_P3213	0x3213
#define	SMT_P3214	0x3214
#define	SMT_P3215	0x3215
#define	SMT_P3216	0x3216
#define	SMT_P3217	0x3217

#define	SMT_P400A	0x400a
#define	SMT_P400B	0x400b
#define	SMT_P400C	0x400c
#define	SMT_P400D	0x400d
#define	SMT_P400E	0x400e
#define	SMT_P400F	0x400f
#define	SMT_P4010	0x4010
#define	SMT_P4011	0x4011
#define	SMT_P4012	0x4012
#define	SMT_P4013	0x4013
#define	SMT_P4014	0x4014
#define	SMT_P4015	0x4015
#define	SMT_P4016	0x4016
#define	SMT_P4017	0x4017
#define	SMT_P4018	0x4018
#define	SMT_P4019	0x4019
#define	SMT_P401A	0x401a
#define	SMT_P401B	0x401b
#define	SMT_P401C	0x401c
#define	SMT_P401D	0x401d
#define	SMT_P401E	0x401e
#define	SMT_P401F	0x401f
#define	SMT_P4020	0x4020
#define	SMT_P4021	0x4021
#define	SMT_P4022	0x4022
#define	SMT_P4023	0x4023
#define	SMT_P4024	0x4024
#define	SMT_P4025	0x4025
#define	SMT_P4026	0x4026
#define	SMT_P4027	0x4027
#define	SMT_P4028	0x4028
#define	SMT_P4029	0x4029
#define	SMT_P402A	0x402a
#define	SMT_P402B	0x402b
#define	SMT_P402C	0x402c
#define	SMT_P402D	0x402d
#define	SMT_P402E	0x402e
#define	SMT_P402F	0x402f
#define	SMT_P4030	0x4030
#define	SMT_P4031	0x4031
#define	SMT_P4032	0x4032
#define	SMT_P4033	0x4033
#define	SMT_P4034	0x4034
#define	SMT_P4035	0x4035
#define	SMT_P4036	0x4036
#define	SMT_P4037	0x4037
#define	SMT_P4038	0x4038
#define	SMT_P4039	0x4039
#define	SMT_P403A	0x403a
#define	SMT_P403B	0x403b
#define	SMT_P403C	0x403c
#define	SMT_P403D	0x403d
#define	SMT_P403E	0x403e
#define	SMT_P403F	0x403f
#define	SMT_P4040	0x4040
#define	SMT_P4041	0x4041
#define	SMT_P4042	0x4042
#define	SMT_P4043	0x4043
#define	SMT_P4044	0x4044
#define	SMT_P4045	0x4045
#define	SMT_P4046	0x4046

#define	SMT_P4050	0x4050
#define	SMT_P4051	0x4051
#define	SMT_P4052	0x4052
#define	SMT_P4053	0x4053
