/*
 * WARNING! This file is automatically generated - DO NOT EDIT!
 */

#define TS_MAGICKEY	0x5a5a5a5a
#define TASK_STATE 0
#define TASK_FLAGS 4
#define TASK_SIGPENDING 8
#define TASK_NEEDRESCHED 20
#define TASK_TSS 470
#define TASK_MM 826
#define TSS_KSP 0
#define TSS_USP 4
#define TSS_SR 8
#define TSS_FS 10
#define TSS_CRP 12
#define TSS_ESP0 20
#define TSS_FPREG 24
#define TSS_FPCNTL 120
#define TSS_FPSTATE 132
#define PT_D0 32
#define PT_ORIG_D0 36
#define PT_D1 0
#define PT_D2 4
#define PT_D3 8
#define PT_D4 12
#define PT_D5 16
#define PT_A0 20
#define PT_A1 24
#define PT_A2 28
#define PT_PC 46
#define PT_SR 44
#define PT_VECTOR 50
#define IRQ_HANDLER 0
#define IRQ_DEVID 8
#define IRQ_NEXT 16
#define STAT_IRQ 120
#define BIR_TAG 0
#define BIR_SIZE 2
#define BIR_DATA 4
#define FBCON_FONT_DESC_IDX 0
#define FBCON_FONT_DESC_NAME 4
#define FBCON_FONT_DESC_WIDTH 8
#define FBCON_FONT_DESC_HEIGHT 12
#define FBCON_FONT_DESC_DATA 16
#define FBCON_FONT_DESC_PREF 20
#define SIGSEGV 11
#define SEGV_MAPERR 1
#define SIGTRAP 5
#define TRAP_TRACE 2
#define CUSTOMBASE -2132807680
#define C_INTENAR 28
#define C_INTREQR 30
#define C_INTENA 154
#define C_INTREQ 156
#define C_SERDATR 24
#define C_SERDAT 48
#define C_SERPER 50
#define CIAABASE -2134908927
#define CIABBASE -2134913024
#define C_PRA 0
#define ZTWOBASE -2147483648
