#undef A_dsa_cmnd
#undef A_dsa_datain
#undef A_dsa_dataout
#undef A_dsa_msgin
#undef A_dsa_msgout
#undef A_dsa_select
#undef A_dsa_size
#undef A_dsa_status
#undef A_had_cmdout
#undef A_had_datain
#undef A_had_dataout
#undef A_had_extmsg
#undef A_had_msgin
#undef A_had_msgout
#undef A_had_select
#undef A_had_status
#undef A_int_bad_extmsg1a
#undef A_int_bad_extmsg1b
#undef A_int_bad_extmsg2a
#undef A_int_bad_extmsg2b
#undef A_int_bad_extmsg3a
#undef A_int_bad_extmsg3b
#undef A_int_bad_msg1
#undef A_int_bad_msg2
#undef A_int_bad_msg3
#undef A_int_cmd_bad_phase
#undef A_int_cmd_complete
#undef A_int_data_bad_phase
#undef A_int_disc1
#undef A_int_disc2
#undef A_int_disc3
#undef A_int_msg_sdtr1
#undef A_int_msg_sdtr2
#undef A_int_msg_sdtr3
#undef A_int_no_msgout1
#undef A_int_no_msgout2
#undef A_int_no_msgout3
#undef A_int_not_cmd_complete
#undef A_int_not_rej
#undef A_int_resel_not_msgin
#undef A_int_reselected
#undef A_int_sel_no_ident
#undef A_int_sel_not_cmd
#undef A_int_selected
#undef A_int_status_not_msgin
#undef A_msgin_buf
#undef A_reselected_identify
#undef Ent_do_select
#undef Ent_done_ident
#undef Ent_end_data_trans
#undef Ent_patch_input_data
#undef Ent_patch_new_dsa
#undef Ent_patch_output_data
#undef Ent_reselect
#undef Ent_resume_cmd
#undef Ent_resume_msgin1a
#undef Ent_resume_msgin1b
#undef Ent_resume_msgin2a
#undef Ent_resume_msgin2b
#undef Ent_resume_msgin3a
#undef Ent_resume_msgin3b
#undef Ent_resume_pmm
#undef Ent_resume_rej_ident
#undef Ent_wait_disc1
#undef Ent_wait_disc2
#undef Ent_wait_disc3
#undef Ent_wait_disc_complete
